@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  * Copyright © 2013 John Morrissey <jwm@horde.net>
  *
@@ -56,6 +56,7 @@
 #include <gnutls/gnutls.h>
 #include <gnutls/abstract.h>
 #include <gnutls/x509.h>
+#include <gnutls/crypto.h>
 #ifdef HAVE_TROUSERS
 #include <trousers/tss.h>
 #include <trousers/trousers.h>
@@ -120,7 +121,22 @@
 struct pkt {
 	int len;
 	struct pkt *next;
-	unsigned char hdr[8];
+	union {
+		struct {
+			uint32_t spi;
+			uint32_t seq;
+			unsigned char iv[16];
+			unsigned char payload[];
+		} esp;
+		struct {
+			unsigned char pad[2];
+			unsigned char hdr[22];
+		} oncp;
+		struct {
+			unsigned char pad[16];
+			unsigned char hdr[8];
+		} cstp;
+	};
 	unsigned char data[];
 };
 
@@ -135,13 +151,22 @@ struct pkt {
 #define KA_REKEY	4
 
 #define DTLS_NOSECRET	0
-#define DTLS_DISABLED	1
-#define DTLS_SLEEPING	2
-#define DTLS_CONNECTING	3
-#define DTLS_CONNECTED	4
+#define DTLS_SECRET	1
+#define DTLS_DISABLED	2
+#define DTLS_SLEEPING	3	/* For ESP, sometimes sending probes */
+#define DTLS_CONNECTING	4	/* ESP probe received; must tell server */
+#define DTLS_CONNECTED	5	/* Server informed and should be sending ESP */
 
 #define COMPR_DEFLATE	(1<<0)
-#define COMPR_ALL	(COMPR_DEFLATE)
+#define COMPR_LZS	(1<<1)
+#define COMPR_LZ4	(1<<2)
+
+#ifdef HAVE_LZ4
+#define COMPR_STATELESS	(COMPR_LZS | COMPR_LZ4)
+#else
+#define COMPR_STATELESS	(COMPR_LZS)
+#endif
+#define COMPR_ALL	(COMPR_STATELESS | COMPR_DEFLATE)
 
 struct keepalive_info {
 	int dpd;
@@ -181,6 +206,7 @@ struct oc_text_buf {
 
 #define MAX_AUTH_TYPES		4
 
+#define AUTH_DEFAULT_DISABLED	-3
 #define AUTH_DISABLED		-2
 #define AUTH_FAILED		-1	/* Failed */
 #define AUTH_UNSEEN		0	/* Server has not offered it */
@@ -188,12 +214,118 @@ struct oc_text_buf {
 	/* Individual auth types may use 2 onwards for their own state */
 #define AUTH_IN_PROGRESS	2	/* In-progress attempt */
 
-struct proxy_auth_state {
+struct http_auth_state {
 	int state;
 	char *challenge;
+	union {
+#ifdef HAVE_GSSAPI
+		struct {
+			gss_name_t gss_target_name;
+			gss_ctx_id_t gss_context;
+		};
+#endif
+#ifdef _WIN32
+		struct {
+			CredHandle ntlm_sspi_cred;
+			CtxtHandle ntlm_sspi_ctx;
+		};
+		struct {
+			CredHandle sspi_cred;
+			CtxtHandle sspi_ctx;
+			SEC_WCHAR *sspi_target_name;
+		};
+#else
+		struct {
+			int ntlm_helper_fd;
+		};
+#endif
+	};
+};
+
+struct vpn_proto {
+	int (*vpn_close_session)(struct openconnect_info *vpninfo, const char *reason);
+
+	/* This does the full authentication, calling back as appropriate */
+	int (*obtain_cookie)(struct openconnect_info *vpninfo);
+
+	/* Establish the TCP connection (and obtain configuration) */
+	int (*tcp_connect)(struct openconnect_info *vpninfo);
+
+	int (*tcp_mainloop)(struct openconnect_info *vpninfo, int *timeout);
+
+	/* Add headers common to each HTTP request */
+	void (*add_http_headers)(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
+
+	/* Set up the UDP (DTLS) connection. Doesn't actually *start* it. */
+	int (*udp_setup)(struct openconnect_info *vpninfo, int attempt_period);
+
+	/* This will actually complete the UDP connection setup/handshake on the wire,
+	   as well as transporting packets */
+	int (*udp_mainloop)(struct openconnect_info *vpninfo, int *timeout);
+
+	/* Close the connection but leave the session setup so it restarts */
+	void (*udp_close)(struct openconnect_info *vpninfo);
+
+	/* Close and destroy the (UDP) session */
+	void (*udp_shutdown)(struct openconnect_info *vpninfo);
+};
+
+struct pkt_q {
+	struct pkt *head;
+	struct pkt **tail;
+	int count;
+};
+
+static inline struct pkt *dequeue_packet(struct pkt_q *q)
+{
+	struct pkt *ret = q->head;
+
+	if (ret) {
+		q->head = ret->next;
+		if (!--q->count)
+			q->tail = &q->head;
+	}
+	return ret;
+}
+
+static inline void requeue_packet(struct pkt_q *q, struct pkt *p)
+{
+	p->next = q->head;
+	q->head = p;
+	if (!q->count++)
+		q->tail = &p->next;
+}
+
+static inline int queue_packet(struct pkt_q *q, struct pkt *p)
+{
+	*(q->tail) = p;
+	p->next = NULL;
+	q->tail = &p->next;
+	return ++q->count;
+}
+
+static inline void init_pkt_queue(struct pkt_q *q)
+{
+	q->tail = &q->head;
+}
+	
+struct esp {
+#if defined(ESP_GNUTLS)
+	gnutls_cipher_hd_t cipher;
+	gnutls_hmac_hd_t hmac;
+#elif defined(ESP_OPENSSL)
+	HMAC_CTX hmac;
+	EVP_CIPHER_CTX cipher;
+#endif
+	uint32_t seq;
+	uint32_t seq_backlog;
+	uint32_t spi; /* Stored network-endian */
+	unsigned char secrets[0x40];
 };
 
 struct openconnect_info {
+	struct vpn_proto proto;
+
 #ifdef HAVE_ICONV
 	iconv_t ic_legacy_to_utf8;
 	iconv_t ic_utf8_to_legacy;
@@ -201,6 +333,19 @@ struct openconnect_info {
 	char *redirect_url;
 	int redirect_type;
 
+	unsigned char esp_hmac;
+	unsigned char esp_enc;
+	unsigned char esp_compr;
+	uint32_t esp_replay_protect;
+	uint32_t esp_lifetime_bytes;
+	uint32_t esp_lifetime_seconds;
+	uint32_t esp_ssl_fallback;
+	int current_esp_in;
+	int old_esp_maxseq;
+	struct esp esp_in[2];
+	struct esp esp_out;
+
+	int tncc_fd; /* For Juniper TNCC */
 	const char *csd_xmltag;
 	int csd_nostub;
 	char *platname;
@@ -230,21 +375,10 @@ struct openconnect_info {
 	char *proxy_user;
 	char *proxy_pass;
 	int proxy_close_during_auth;
-	struct proxy_auth_state auth[MAX_AUTH_TYPES];
-#ifdef HAVE_GSSAPI
-	gss_name_t gss_target_name;
-	gss_ctx_id_t gss_context;
-#endif
-#ifdef _WIN32
-	CredHandle ntlm_sspi_cred;
-	CtxtHandle ntlm_sspi_ctx;
-	CredHandle sspi_cred;
-	CtxtHandle sspi_ctx;
-	SEC_WCHAR *sspi_target_name;
-#else
-	int ntlm_helper_fd;
-#endif
-	int authmethods_set;
+	int retry_on_auth_fail;
+	int try_http_auth;
+	struct http_auth_state http_auth[MAX_AUTH_TYPES];
+	struct http_auth_state proxy_auth[MAX_AUTH_TYPES];
 
 	char *localname;
 	char *hostname;
@@ -283,16 +417,19 @@ struct openconnect_info {
 	pskc_t *pskc;
 	pskc_key_t *pskc_key;
 #endif
-#ifdef HAVE_LIBOATH
 	char *oath_secret;
 	size_t oath_secret_len;
+	enum {
+		OATH_ALG_HMAC_SHA1 = 0,
+		OATH_ALG_HMAC_SHA256,
+		OATH_ALG_HMAC_SHA512,
+	} oath_hmac_alg;
 	enum {
 		HOTP_SECRET_BASE32 = 1,
 		HOTP_SECRET_RAW,
 		HOTP_SECRET_HEX,
 		HOTP_SECRET_PSKC,
 	} hotp_secret_format; /* We need to give it back in the same form */
-#endif
 #ifdef HAVE_LIBPCSCLITE
 	SCARDHANDLE pcsc_ctx, pcsc_card;
 	char *yubikey_objname;
@@ -357,12 +494,13 @@ struct openconnect_info {
 	struct pkt *deflate_pkt;		/* For compressing outbound packets into */
 	struct pkt *pending_deflated_pkt;	/* The original packet associated with above */
 	struct pkt *current_ssl_pkt;		/* Partially sent SSL packet */
-
+	struct pkt_q oncp_control_queue;		/* Control packets to be sent on oNCP next */
 	/* Packet buffers for receiving into */
 	struct pkt *cstp_pkt;
 	struct pkt *dtls_pkt;
 	struct pkt *tun_pkt;
-	
+	int pkt_trailer; /* How many bytes after payload for encryption (ESP HMAC) */
+
 	z_stream inflate_strm;
 	uint32_t inflate_adler32;
 	z_stream deflate_strm;
@@ -376,7 +514,6 @@ struct openconnect_info {
 #if defined(DTLS_OPENSSL)
 	SSL_CTX *dtls_ctx;
 	SSL *dtls_ssl;
-	SSL_SESSION *dtls_session;
 #elif defined(DTLS_GNUTLS)
 	/* Call this dtls_ssl rather than dtls_sess because it's just a
 	   pointer, and generic code in dtls.c wants to check if it's
@@ -433,9 +570,8 @@ struct openconnect_info {
 	int got_pause_cmd;
 	char cancel_type;
 
-	struct pkt *incoming_queue;
-	struct pkt *outgoing_queue;
-	int outgoing_qlen;
+	struct pkt_q incoming_queue;
+	struct pkt_q outgoing_queue;
 	int max_qlen;
 	struct oc_stats stats;
 	openconnect_stats_vfn stats_handler;
@@ -455,6 +591,7 @@ struct openconnect_info {
 
 	const char *quit_reason;
 
+	int verbose;
 	void *cbdata;
 	openconnect_validate_peer_cert_vfn validate_peer_cert;
 	openconnect_write_new_config_vfn write_new_config;
@@ -509,7 +646,10 @@ struct openconnect_info {
 #define AC_PKT_COMPRESSED	8	/* Compressed data */
 #define AC_PKT_TERM_SERVER	9	/* Server kick */
 
-#define vpn_progress(vpninfo, ...) (vpninfo)->progress((vpninfo)->cbdata, __VA_ARGS__)
+#define vpn_progress(_v, lvl, ...) do {					\
+	if ((_v)->verbose >= (lvl))					\
+		(_v)->progress((_v)->cbdata, lvl, __VA_ARGS__);	\
+	} while(0)
 #define vpn_perror(vpninfo, msg) vpn_progress((vpninfo), PRG_ERR, "%s: %s\n", (msg), strerror(errno));
 
 /****************************************************************************/
@@ -606,11 +746,13 @@ char *openconnect_legacy_to_utf8(struct openconnect_info *vpninfo, const char *l
 #endif
 
 /* script.c */
+unsigned char unhex(const char *data);
 int script_setenv(struct openconnect_info *vpninfo, const char *opt, const char *val, int append);
 int script_setenv_int(struct openconnect_info *vpninfo, const char *opt, int value);
 void prepare_script_env(struct openconnect_info *vpninfo);
 int script_config_tun(struct openconnect_info *vpninfo, const char *reason);
 int apply_script_env(struct oc_vpn_option *envs);
+void free_split_routes(struct openconnect_info *vpninfo);
 
 /* tun.c / tun-win32.c */
 void os_shutdown_tun(struct openconnect_info *vpninfo);
@@ -619,17 +761,32 @@ int os_write_tun(struct openconnect_info *vpninfo, struct pkt *pkt);
 intptr_t os_setup_tun(struct openconnect_info *vpninfo);
 
 /* dtls.c */
-unsigned char unhex(const char *data);
+int dtls_setup(struct openconnect_info *vpninfo, int dtls_attempt_period);
 int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout);
-int dtls_try_handshake(struct openconnect_info *vpninfo);
-int connect_dtls_socket(struct openconnect_info *vpninfo);
 void dtls_close(struct openconnect_info *vpninfo);
 void dtls_shutdown(struct openconnect_info *vpninfo);
 
 /* cstp.c */
+void cstp_common_headers(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
+int cstp_connect(struct openconnect_info *vpninfo);
 int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout);
 int cstp_bye(struct openconnect_info *vpninfo, const char *reason);
-void cstp_free_splits(struct openconnect_info *vpninfo);
+int decompress_and_queue_packet(struct openconnect_info *vpninfo, int compr_type,
+				unsigned char *buf, int len);
+int compress_packet(struct openconnect_info *vpninfo, int compr_type, struct pkt *this);
+
+/* auth-juniper.c */
+int oncp_obtain_cookie(struct openconnect_info *vpninfo);
+void oncp_common_headers(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
+
+/* oncp.c */
+int queue_esp_control(struct openconnect_info *vpninfo, int enable);
+int oncp_connect(struct openconnect_info *vpninfo);
+int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout);
+
+/* lzs.c */
+int lzs_decompress(unsigned char *dst, int dstlen, const unsigned char *src, int srclen);
+int lzs_compress(unsigned char *dst, int dstlen, const unsigned char *src, int srclen);
 
 /* ssl.c */
 unsigned string_is_hostname(const char* str);
@@ -656,14 +813,33 @@ int openconnect_open_utf8(struct openconnect_info *vpninfo,
 			  const char *fname, int mode);
 FILE *openconnect_fopen_utf8(struct openconnect_info *vpninfo,
 			     const char *fname, const char *mode);
-
+int udp_sockaddr(struct openconnect_info *vpninfo, int port);
+int udp_connect(struct openconnect_info *vpninfo);
+int ssl_reconnect(struct openconnect_info *vpninfo);
 void openconnect_clear_cookies(struct openconnect_info *vpninfo);
 
 /* openssl-pkcs11.c */
 int load_pkcs11_key(struct openconnect_info *vpninfo);
 int load_pkcs11_certificate(struct openconnect_info *vpninfo);
 
+/* esp.c */
+int verify_packet_seqno(struct openconnect_info *vpninfo,
+			struct esp *esp, uint32_t seq);
+int esp_setup(struct openconnect_info *vpninfo, int dtls_attempt_period);
+int esp_mainloop(struct openconnect_info *vpninfo, int *timeout);
+void esp_close(struct openconnect_info *vpninfo);
+void esp_shutdown(struct openconnect_info *vpninfo);
+int print_esp_keys(struct openconnect_info *vpninfo, const char *name, struct esp *esp);
+
+/* {gnutls,openssl}-esp.c */
+int setup_esp_keys(struct openconnect_info *vpninfo);
+void destroy_esp_ciphers(struct esp *esp);
+int decrypt_esp_packet(struct openconnect_info *vpninfo, struct esp *esp, struct pkt *pkt);
+int encrypt_esp_packet(struct openconnect_info *vpninfo, struct pkt *pkt);
+
 /* {gnutls,openssl}.c */
+int ssl_nonblock_read(struct openconnect_info *vpninfo, void *buf, int maxlen);
+int ssl_nonblock_write(struct openconnect_info *vpninfo, void *buf, int buflen);
 int openconnect_open_https(struct openconnect_info *vpninfo);
 void openconnect_close_https(struct openconnect_info *vpninfo, int final);
 int cstp_handshake(struct openconnect_info *vpninfo, unsigned init);
@@ -677,7 +853,9 @@ int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
 int openconnect_yubikey_chalresp(struct openconnect_info *vpninfo,
 				 const void *challenge, int chall_len, void *result);
 int openconnect_hash_yubikey_password(struct openconnect_info *vpninfo,
-				      const char *password, const void *ident, int id_len);
+				      const char *password, int pwlen,
+				      const void *ident, int id_len);
+int hotp_hmac(struct openconnect_info *vpninfo, const void *challenge);
 #if defined(OPENCONNECT_OPENSSL)
 #define openconnect_https_connected(_v) ((_v)->https_ssl)
 #elif defined (OPENCONNECT_GNUTLS)
@@ -686,8 +864,7 @@ int openconnect_hash_yubikey_password(struct openconnect_info *vpninfo,
 
 /* mainloop.c */
 int tun_mainloop(struct openconnect_info *vpninfo, int *timeout);
-int queue_new_packet(struct pkt **q, void *buf, int len);
-void queue_packet(struct pkt **q, struct pkt *new);
+int queue_new_packet(struct pkt_q *q, void *buf, int len);
 int keepalive_action(struct keepalive_info *ka, int *timeout);
 int ka_stalled_action(struct keepalive_info *ka, int *timeout);
 
@@ -732,54 +909,90 @@ int do_gen_yubikey_code(struct openconnect_info *vpninfo,
 			struct oc_form_opt *opt);
 
 /* auth.c */
-void nuke_opt_values(struct oc_form_opt *opt);
-int parse_xml_response(struct openconnect_info *vpninfo, char *response,
-		       struct oc_auth_form **form, int *cert_rq);
-int process_auth_form(struct openconnect_info *vpninfo,
-		      struct oc_auth_form *form);
-int handle_auth_form(struct openconnect_info *vpninfo,
-		     struct oc_auth_form *form,
-		     struct oc_text_buf *request_body, const char **method,
-		     const char **request_body_type);
+int cstp_obtain_cookie(struct openconnect_info *vpninfo);
+
+/* auth-common.c */
+int xmlnode_is_named(xmlNode *xml_node, const char *name);
+int xmlnode_get_prop(xmlNode *xml_node, const char *name, char **var);
+int xmlnode_match_prop(xmlNode *xml_node, const char *name, const char *match);
+int append_opt(struct oc_text_buf *body, char *opt, char *name);
+int append_form_opts(struct openconnect_info *vpninfo,
+		     struct oc_auth_form *form, struct oc_text_buf *body);
+void free_opt(struct oc_form_opt *opt);
 void free_auth_form(struct oc_auth_form *form);
-int xmlpost_initial_req(struct openconnect_info *vpninfo,
-			struct oc_text_buf *request_body, int cert_fail);
+int do_gen_tokencode(struct openconnect_info *vpninfo,
+		     struct oc_auth_form *form);
+int can_gen_tokencode(struct openconnect_info *vpninfo,
+		      struct oc_auth_form *form,
+		      struct oc_form_opt *opt);
 
 /* http.c */
 struct oc_text_buf *buf_alloc(void);
+void dump_buf(struct openconnect_info *vpninfo, char prefix, char *buf);
 int buf_ensure_space(struct oc_text_buf *buf, int len);
 void  __attribute__ ((format (printf, 2, 3)))
 	buf_append(struct oc_text_buf *buf, const char *fmt, ...);
 void buf_append_bytes(struct oc_text_buf *buf, const void *bytes, int len);
-void buf_append_base64(struct oc_text_buf *buf, const void *bytes, int len);
 int buf_append_utf16le(struct oc_text_buf *buf, const char *utf8);
+int get_utf8char(const char **utf8);
 void buf_append_from_utf16le(struct oc_text_buf *buf, const void *utf16);
-void *openconnect_base64_decode(int *len, const char *in);
 void buf_truncate(struct oc_text_buf *buf);
 void buf_append_urlencoded(struct oc_text_buf *buf, char *str);
 int buf_error(struct oc_text_buf *buf);
 int buf_free(struct oc_text_buf *buf);
 char *openconnect_create_useragent(const char *base);
-void cleanup_proxy_auth(struct openconnect_info *vpninfo);
 int process_proxy(struct openconnect_info *vpninfo, int ssl_sock);
 int internal_parse_url(const char *url, char **res_proto, char **res_host,
 		       int *res_port, char **res_path, int default_port);
+int do_https_request(struct openconnect_info *vpninfo, const char *method,
+		     const char *request_body_type, struct oc_text_buf *request_body,
+		     char **form_buf, int fetch_redirect);
+int http_add_cookie(struct openconnect_info *vpninfo, const char *option,
+		    const char *value, int replace);
+int process_http_response(struct openconnect_info *vpninfo, int connect,
+			  int (*header_cb)(struct openconnect_info *, char *, char *),
+			  struct oc_text_buf *body);
+int handle_redirect(struct openconnect_info *vpninfo);
+void http_common_headers(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
 
+/* http-auth.c */
+void buf_append_base64(struct oc_text_buf *buf, const void *bytes, int len);
+void *openconnect_base64_decode(int *len, const char *in);
+void clear_auth_states(struct openconnect_info *vpninfo,
+		       struct http_auth_state *auth_states, int reset);
+int proxy_auth_hdrs(struct openconnect_info *vpninfo, char *hdr, char *val);
+int http_auth_hdrs(struct openconnect_info *vpninfo, char *hdr, char *val);
+int gen_authorization_hdr(struct openconnect_info *vpninfo, int proxy,
+			  struct oc_text_buf *buf);
 /* ntlm.c */
-int ntlm_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
-void cleanup_ntlm_auth(struct openconnect_info *vpninfo);
+int ntlm_authorization(struct openconnect_info *vpninfo, int proxy, struct http_auth_state *auth_state, struct oc_text_buf *buf);
+void cleanup_ntlm_auth(struct openconnect_info *vpninfo, struct http_auth_state *auth_state);
 
 /* gssapi.c */
-int gssapi_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
-void cleanup_gssapi_auth(struct openconnect_info *vpninfo);
+int gssapi_authorization(struct openconnect_info *vpninfo, int proxy, struct http_auth_state *auth_state, struct oc_text_buf *buf);
+void cleanup_gssapi_auth(struct openconnect_info *vpninfo, struct http_auth_state *auth_state);
 int socks_gssapi_auth(struct openconnect_info *vpninfo);
 
 /* digest.c */
-int digest_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
+int digest_authorization(struct openconnect_info *vpninfo, int proxy, struct http_auth_state *auth_state, struct oc_text_buf *buf);
+
+/* library.c */
+void nuke_opt_values(struct oc_form_opt *opt);
+int process_auth_form(struct openconnect_info *vpninfo, struct oc_auth_form *form);
+/* This is private for now since we haven't yet worked out what the API will be */
+void openconnect_set_juniper(struct openconnect_info *vpninfo);
 
 /* version.c */
 extern const char *openconnect_version_str;
 
+/* strncasecmp() just checks that the first n characters match. This
+   function ensures that the first n characters of the left-hand side
+   are a *precise* match for the right-hand side. */
+static inline int strprefix_match(const char *str, int len, const char *match)
+{
+	return len == strlen(match) && !strncasecmp(str, match, len);
+}
+
 #define STRDUP(res, arg) \
 	do {							\
 		free(res);					\
@@ -805,4 +1018,111 @@ extern const char *openconnect_version_str;
 		return;							\
 	}
 
+/* Let's stop open-coding big-endian and little-endian loads/stores.
+ *
+ * Start with a packed structure so that we can let the compiler
+ * decide whether the target CPU can cope with unaligned load/stores
+ * or not. Then there are three cases to handle:
+ *  - For big-endian loads/stores, just use htons() et al.
+ *  - For little-endian when we *know* the CPU is LE, just load/store
+ *  - For little-endian otherwise, do the data acess byte-wise
+ */
+struct oc_packed_uint32_t {
+	uint32_t d;
+} __attribute__((packed));
+struct oc_packed_uint16_t {
+	uint16_t d;
+} __attribute__((packed));
+
+static inline uint32_t load_be32(const void *_p)
+{
+	const struct oc_packed_uint32_t *p = _p;
+	return ntohl(p->d);
+}
+
+static inline uint16_t load_be16(const void *_p)
+{
+	const struct oc_packed_uint16_t *p = _p;
+	return ntohs(p->d);
+}
+
+static inline void store_be32(void *_p, uint32_t d)
+{
+	struct oc_packed_uint32_t *p = _p;
+	p->d = htonl(d);
+}
+
+static inline void store_be16(void *_p, uint16_t d)
+{
+	struct oc_packed_uint16_t *p = _p;
+	p->d = htons(d);
+}
+
+/* It doesn't matter if we don't find one. It'll default to the
+ * "not known to be little-endian" case, and do the bytewise
+ * load/store. Modern compilers might even spot the pattern and
+ * optimise it (see GCC PR#55177 around comment 15). */
+#ifdef ENDIAN_HDR
+#include ENDIAN_HDR
+#endif
+
+#if defined(_WIN32) ||							       \
+   (defined(_LITTLE_ENDIAN) && !defined(_BIG_ENDIAN)) /* Solaris */ ||	       \
+   (defined(__LITTLE_ENDIAN) && defined(__BIG_ENDIAN) && defined(__BYTE_ORDER) \
+    && __BYTE_ORDER == __LITTLE_ENDIAN) /* Linux */ ||			       \
+   (defined(LITTLE_ENDIAN) && defined(BIG_ENDIAN) && defined(BYTE_ORDER)       \
+    && BYTE_ORDER == LITTLE_ENDIAN) /* *BSD */
+static inline uint32_t load_le32(const void *_p)
+{
+	const struct oc_packed_uint32_t *p = _p;
+	return p->d;
+}
+
+static inline uint16_t load_le16(const void *_p)
+{
+	const struct oc_packed_uint16_t *p = _p;
+	return p->d;
+}
+
+static inline void store_le32(void *_p, uint32_t d)
+{
+	struct oc_packed_uint32_t *p = _p;
+	p->d = d;
+}
+
+static inline void store_le16(void *_p, uint16_t d)
+{
+	struct oc_packed_uint16_t *p = _p;
+	p->d = d;
+}
+#else
+static inline uint32_t load_le32(const void *_p)
+{
+	const unsigned char *p = _p;
+	return p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24);
+}
+
+static inline uint16_t load_le16(const void *_p)
+{
+	const unsigned char *p = _p;
+	return p[0] | (p[1] << 8);
+}
+
+static inline void store_le32(void *_p, uint32_t d)
+{
+	unsigned char *p = _p;
+	p[0] = d;
+	p[1] = d >> 8;
+}
+
+static inline void store_le16(void *_p, uint16_t d)
+{
+	unsigned char *p = _p;
+	p[0] = d;
+	p[1] = d >> 8;
+	p[2] = d >> 16;
+	p[3] = d >> 24;
+}
+#endif /* !Not known to be little-endian */
+
 #endif /* __OPENCONNECT_INTERNAL_H__ */
