@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -369,7 +369,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
  out:
 	/* If proxy processing returned -EAGAIN to reconnect before attempting
 	   further auth, and we failed to reconnect, we have to clean up here. */
-	cleanup_proxy_auth(vpninfo);
+	clear_auth_states(vpninfo, vpninfo->proxy_auth, 1);
 	return ssl_sock;
 }
 
@@ -622,8 +622,7 @@ int keystore_fetch(const char *key, unsigned char **result)
 	}
 	len = strlen(key);
 	buf[0] = 'g';
-	buf[1] = len >> 8;
-	buf[2] = len & 0xff;
+	store_be16(buf + 1, len);
 
 	if (send(fd, buf, 3, 0) != 3 || send(fd, key, len, 0) != len ||
 	    shutdown(fd, SHUT_WR) || recv(fd, buf, 1, 0) != 1)
@@ -636,7 +635,7 @@ int keystore_fetch(const char *key, unsigned char **result)
 	}
 	if (recv(fd, buf, 2, 0) != 2)
 		goto out;
-	len = (buf[0] << 8) + buf[1];
+	len = load_be16(buf);
 	data = malloc(len);
 	if (!data)
 		goto out;
@@ -798,3 +797,132 @@ FILE *openconnect_fopen_utf8(struct openconnect_info *vpninfo, const char *fname
 
 	return fdopen(fd, mode);
 }
+
+int udp_sockaddr(struct openconnect_info *vpninfo, int port)
+{
+	vpninfo->dtls_addr = malloc(vpninfo->peer_addrlen);
+	if (!vpninfo->dtls_addr)
+		return -ENOMEM;
+
+	memcpy(vpninfo->dtls_addr, vpninfo->peer_addr, vpninfo->peer_addrlen);
+
+	if (vpninfo->peer_addr->sa_family == AF_INET) {
+		struct sockaddr_in *sin = (void *)vpninfo->dtls_addr;
+		sin->sin_port = htons(port);
+	} else if (vpninfo->peer_addr->sa_family == AF_INET6) {
+		struct sockaddr_in6 *sin = (void *)vpninfo->dtls_addr;
+		sin->sin6_port = htons(port);
+	} else {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unknown protocol family %d. Cannot create UDP server address\n"),
+			     vpninfo->peer_addr->sa_family);
+		return -EINVAL;
+	}
+
+	return 0;
+}
+
+int udp_connect(struct openconnect_info *vpninfo)
+{
+	int fd, sndbuf;
+
+	fd = socket(vpninfo->peer_addr->sa_family, SOCK_DGRAM, IPPROTO_UDP);
+	if (fd < 0) {
+		vpn_perror(vpninfo, _("Open UDP socket"));
+		return -EINVAL;
+	}
+	if (vpninfo->protect_socket)
+		vpninfo->protect_socket(vpninfo->cbdata, fd);
+
+	sndbuf = vpninfo->ip_info.mtu * 2;
+	setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (void *)&sndbuf, sizeof(sndbuf));
+
+	if (vpninfo->dtls_local_port) {
+		union {
+			struct sockaddr_in in;
+			struct sockaddr_in6 in6;
+		} dtls_bind_addr;
+		int dtls_bind_addrlen;
+		memset(&dtls_bind_addr, 0, sizeof(dtls_bind_addr));
+
+		if (vpninfo->peer_addr->sa_family == AF_INET) {
+			struct sockaddr_in *addr = &dtls_bind_addr.in;
+			dtls_bind_addrlen = sizeof(*addr);
+			addr->sin_family = AF_INET;
+			addr->sin_addr.s_addr = INADDR_ANY;
+			addr->sin_port = htons(vpninfo->dtls_local_port);
+		} else if (vpninfo->peer_addr->sa_family == AF_INET6) {
+			struct sockaddr_in6 *addr = &dtls_bind_addr.in6;
+			dtls_bind_addrlen = sizeof(*addr);
+			addr->sin6_family = AF_INET6;
+			addr->sin6_addr = in6addr_any;
+			addr->sin6_port = htons(vpninfo->dtls_local_port);
+		} else {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Unknown protocol family %d. Cannot use UDP transport\n"),
+				     vpninfo->peer_addr->sa_family);
+			vpninfo->dtls_attempt_period = 0;
+			closesocket(fd);
+			return -EINVAL;
+		}
+
+		if (bind(fd, (struct sockaddr *)&dtls_bind_addr, dtls_bind_addrlen)) {
+			vpn_perror(vpninfo, _("Bind UDP socket"));
+			closesocket(fd);
+			return -EINVAL;
+		}
+	}
+
+	if (connect(fd, vpninfo->dtls_addr, vpninfo->peer_addrlen)) {
+		vpn_perror(vpninfo, _("Connect UDP socket\n"));
+		closesocket(fd);
+		return -EINVAL;
+	}
+
+	set_fd_cloexec(fd);
+	set_sock_nonblock(fd);
+
+	return fd;
+}
+
+int ssl_reconnect(struct openconnect_info *vpninfo)
+{
+	int ret;
+	int timeout;
+	int interval;
+
+	openconnect_close_https(vpninfo, 0);
+
+
+	timeout = vpninfo->reconnect_timeout;
+	interval = vpninfo->reconnect_interval;
+
+	free(vpninfo->dtls_pkt);
+	vpninfo->dtls_pkt = NULL;
+	free(vpninfo->tun_pkt);
+	vpninfo->tun_pkt = NULL;
+
+	while ((ret = vpninfo->proto.tcp_connect(vpninfo))) {
+		if (timeout <= 0)
+			return ret;
+		if (ret == -EPERM) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Cookie is no longer valid, ending session\n"));
+			return ret;
+		}
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("sleep %ds, remaining timeout %ds\n"),
+			     interval, timeout);
+		poll_cmd_fd(vpninfo, interval);
+		if (vpninfo->got_cancel_cmd)
+			return -EINTR;
+		if (vpninfo->got_pause_cmd)
+			return 0;
+		timeout -= interval;
+		interval += vpninfo->reconnect_interval;
+		if (interval > RECONNECT_INTERVAL_MAX)
+			interval = RECONNECT_INTERVAL_MAX;
+	}
+	script_config_tun(vpninfo, "reconnect");
+	return 0;
+}
